@@ -19,17 +19,25 @@ return {
     'BufWriteCmd',            -- write buffer using command
     'BufWritePost',           -- after writing a buffer
     'BufWritePre',            -- before writing a buffer
+    'ChanInfo',               -- info was received about channel
+    'ChanOpen',               -- channel was opened
+    'CmdLineChanged',         -- command line was modified
+    'CmdLineEnter',           -- after entering cmdline mode
+    'CmdLineLeave',           -- before leaving cmdline mode
     'CmdUndefined',           -- command undefined
     'CmdWinEnter',            -- after entering the cmdline window
     'CmdWinLeave',            -- before leaving the cmdline window
     'ColorScheme',            -- after loading a colorscheme
+    'ColorSchemePre',         -- before loading a colorscheme
     'CompleteDone',           -- after finishing insert complete
     'CursorHold',             -- cursor in same position for a while
     'CursorHoldI',            -- idem, in Insert mode
     'CursorMoved',            -- cursor was moved
     'CursorMovedI',           -- cursor was moved in Insert mode
+    'DiffUpdated',            -- diffs have been updated
     'DirChanged',             -- directory changed
     'EncodingChanged',        -- after changing the 'encoding' option
+    'ExitPre',                -- before exiting
     'FileAppendCmd',          -- append to a file using command
     'FileAppendPost',         -- after appending to a file
     'FileAppendPre',          -- before appending to a file
@@ -83,16 +91,19 @@ return {
     'TermOpen',               -- after opening a terminal buffer
     'TermResponse',           -- after setting "v:termresponse"
     'TextChanged',            -- text was modified
-    'TextChangedI',           -- text was modified in Insert mode
+    'TextChangedI',           -- text was modified in Insert mode(no popup)
+    'TextChangedP',           -- text was modified in Insert mode(popup)
     'TextYankPost',           -- after a yank or delete was done (y, d, c)
     'User',                   -- user defined autocommand
     'VimEnter',               -- after starting Vim
     'VimLeave',               -- before exiting Vim
     'VimLeavePre',            -- before exiting Vim and writing ShaDa file
     'VimResized',             -- after Vim window was resized
-    'WinNew',                 -- when entering a new window
+    'VimResume',              -- after Nvim is resumed
+    'VimSuspend',             -- before Nvim is suspended
     'WinEnter',               -- after entering a window
     'WinLeave',               -- before leaving a window
+    'WinNew',                 -- when entering a new window
   },
   aliases = {
     BufCreate = 'BufAdd',
@@ -100,9 +111,9 @@ return {
     BufWrite = 'BufWritePre',
     FileEncoding = 'EncodingChanged',
   },
-  -- List of neovim-specific events or aliases for the purpose of generating 
+  -- List of nvim-specific events or aliases for the purpose of generating
   -- syntax file
-  neovim_specific = {
+  nvim_specific = {
     DirChanged=true,
     TabClosed=true,
     TabNew=true,
